/*

-Procedure turnpar ( Determine rate prameters of a spacecraft turn. )

   Jeff Bytof, last update 8-June-2001 

*/

#include "SpiceUsr.h"
#include "SpiceZmc.h"
#include "SpiceZfc.h"



void turnpar ( SpiceInt         topt, 
               SpiceDouble      rt1        [3], 
               SpiceDouble      rt2        [3],
               SpiceDouble      t,
               SpiceDouble      totacc,
               SpiceDouble    * tbout,
               SpiceDouble    * tcout,
               SpiceDouble      alpha     [3],
               SpiceDouble      omega     [3],
               SpiceDouble      rc1       [3],
               SpiceDouble      rc2       [3],
               SpiceBoolean   * found,
               SpiceInt       * ier                   
             )

{


/*

-Brief_I/O

    Variable    I/O      Description
    --------    ---      -----------
    topt         I       Input turn vector option. 
                            topt = 1, input vectors rt1 and rt2 point to 
                                      the start and end of the turn. 
                            topt = 2, inputs rt1 and rt2[0] are the turn axis
                                      vector and turn angle. 
    rt1          I       First turn vector. 
    rt2          I       Second turn vector. 
    t            I       Total duration of turn, in seconds.
    totacc       I       Requested magnitude of angular acceleration vector.

    tbout        O       Duration of spinup, in seconds.
    tcout        O       Duration of unaccelerated coast, in seconds.
    alpha        O       Angular acceleration vector (rad/sec^2). 
    omega        O       Angular velocity vector (rad/sec).
    rc1          O       First complement turn vector.
    rc2          O       Second complement turn vector.
    found        O       Logical flag, SPICETRUE if no error is found.
    ier          O       If found = SPICEFALSE returns an error code.

*/
   
/*
   Local variables.
*/
 
   
   SpiceDouble   a;
   SpiceDouble   angvec     [3]; 
   SpiceDouble   b;
   SpiceDouble   c;
   SpiceDouble   disc;
   SpiceDouble   r1         [3];
   SpiceDouble   r2         [3];
   SpiceDouble   root1      [2];
   SpiceDouble   root2      [2];
   SpiceDouble   tb;
   SpiceDouble   tc;
   SpiceDouble   theta;
   SpiceDouble   TOL; 
   SpiceDouble   z          [3]; 
   SpiceDouble   zero       [3]; 
 
   SpiceInt      i;

/*
   Set variables.
*/

   TOL = 1.e-8;

   *found = SPICEFALSE;
   *ier   = 0;

   vpack_c ( 0.0, 0.0, 1.0, z );
   vpack_c ( 0.0, 0.0, 0.0, zero );

   vequ_c ( rt1, r1 );
   vequ_c ( rt2, r2 );


/*
   Check for bad inputs. 
*/

   if ( t < 0.0 ) {
      *ier = 1;
      return;
   }

   if ( totacc < 0.0 ) {
      *ier = 2;
      return;
   }


/*
   Get turn angle.
*/

   if ( topt == 1 ) { 
      theta = vsep_c ( r1, r2 );
   } else if ( topt == 2 ) { 
      theta = r2[0];
   } else { 
      
/*
   Bad turn vector code.
*/

      *ier = 3;
      return;
   }

/*
   If turn angle is nearly zero, return with error.
*/

   if ( theta < TOL ) { 
      *ier = 4;
      return;
   }

/*
   If turn axis/angle option (topt=2) is in effect, perform
   transformation to the complementary turn vectors.
*/

   if ( topt == 2 ) { 

/*
   Generate turn vector r1.
*/

      if ( vprox ( rt1, z, TOL ) == SPICETRUE ) {

/*
   Special case is where rt1 lines up with the Z-axis. Set
   r1 in the direction of the X-axis. 
*/ 
       
         vpack_c ( 1.0, 0.0, 0.0, r1 ); 

      } else { 

         ucrss_c ( rt1, z, r1 ); 

      }

/*
   Rotate first vector around the turn axis to get the second vector.
*/

      vrotv_c ( r1, rt1, theta, r2 );

/*
   Store complement vector set.
*/

      vequ_c ( r1, rc1 );
      vequ_c ( r2, rc2 );

   } else if ( topt == 1 ) {

/*
   Get turn/axis/angle complement set.
*/

      ucrss_c ( r1, r2, rc1 );
      vpack_c ( theta, 0.0, 0.0, rc2 );

   }

/*
   Setup quadratic formula to solve for boost duration.
*/

   a = -totacc;
   b =  totacc * t;
   c = -theta;

/*
   Obtain discriminant.
*/

   disc = b*b - 4.0*a*c;

/*
   Error exit on imaginary solutions.
*/

   if ( disc < 0.0 ) {
      *ier = 5;
      return;
   }

/*
   Find and use first root.
*/

   rquad_c ( a, b, c, root1, root2 ); 

   tb = root1[0];

/*
   Verify that boost duration is within bounds. 
*/

   if ( ( tb <= 0.0 ) || ( tb*2.0 > t ) ) {
      *ier = 6;
      return;
   }


/*
   Determine coast duration.
*/

   tc = t - ( tb * 2.0 );

/*
   Obtain direction of angular velocity vector.
*/

   ucrss_c ( r1, r2, angvec );

/*
   Obtain components of angular acceleration vector.
*/
   
   vscl_c ( totacc, angvec, alpha );

/*
   Obtain components of angular velocity vector.
*/

   vscl_c ( tb, alpha, omega );

   *tbout = tb;
   *tcout = tc;
   *found = SPICETRUE;

   return;

}



SpiceBoolean vprox ( SpiceDouble v1  [3], 
                     SpiceDouble v2  [3],
                     SpiceDouble TOL )

/*
   Check for approximate equality of two vectors.
*/

#include "SpiceUsr.h"
#include "SpiceZmc.h"
#include "SpiceZfc.h"

{
   
/*
   Local variables.
*/
   
   SpiceInt     i;

   for ( i=0; i<3; i++ ) { 
      if ( fabs( v1[i]-v2[i] ) < TOL ) {
         return SPICEFALSE;
      }
   }

   return SPICETRUE;
 
}
